/////////////////////////////////////////////////////////////////////////////
//
// Undelete
//
// Copyright (C) 1998 Mark Russinovich and Bryce Cogswell
// Systems Internals - http://www.sysinternals.com
//
// See readme.txt for terms of use.
//
/////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "afxmt.h"
#include "Undelete.h"
#include "BinHeader.h"
#include "winplace.h"
#include "UndelDoc.h"
#include "InfoEntry.h"
#include "Prop.h"
#include "Filter.h"
#include "DriveProp.h"
#include "PropSh.h"
#include "MainFrm.h"
#include "winternals.h"
#include "Fileprop.h"
#include <direct.h>
#include "winioctl.h"
#include "..\dd\ioctlcmd.h"

#include "ListVwEx.h"	// base class for CUndeleteView
#include "UndelVw.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView

IMPLEMENT_DYNCREATE(CUndeleteView, CListViewEx)

BEGIN_MESSAGE_MAP(CUndeleteView, CListViewEx)
	//{{AFX_MSG_MAP(CUndeleteView)
	ON_WM_LBUTTONDBLCLK()
	ON_WM_DESTROY()
	ON_COMMAND(ID_DELETE, OnDelete)
	ON_COMMAND(ID_EMPTYBIN, OnEmptybin)
	ON_COMMAND(ID_UNDELETE, OnUndelete)
	ON_COMMAND(ID_REFRESH, OnRefresh)
	ON_COMMAND(ID_PROPERTIES, OnProperties)
	ON_COMMAND(ID_EXCLUDE, OnExclude)
	ON_WM_CONTEXTMENU()
	ON_WM_ACTIVATE()
	ON_WM_KEYDOWN()
	ON_WM_LBUTTONDOWN()
	ON_WM_HSCROLL()
	ON_COMMAND(ID_FILEPROPERTIES, OnFileProperties)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView construction/destruction

CUndeleteView::CUndeleteView(): m_Mutex( FALSE, _T("UndeleteMutex"))
{
	HKEY hKey;
	DWORD	i;
	DWORD	index;
	DWORD	len;
	DWORD	numFilters, sizeDirFilters;
	PFILTERENTRY filters;
	CString	filterString;
	PCHAR	dirFilters;
	HANDLE  hMutex;
	HANDLE	driverHandle;

	// initialize member variables and read property information
	m_nCheckedItem = -1;
	m_hSmallIcon = NULL;
	m_hLargeIcon = NULL;
	m_GlobalRegistryKey = _T("SOFTWARE\\Systems Internals\\Undelete");
	m_LocalRegistryKey = _T("SOFTWARE\\Systems Internals\\Undelete");
	GetTextualSID();

	// make default global settings
	m_Properties.SizeOfInformation = sizeof(m_Properties);
	m_Properties.EnforceGlobalSetting = 1;
	for( i = 0; i < 28; i++ ) m_Properties.ReservedForRecycle[i] = 10;
	m_Properties.RecycleDisableMask = 0;

	// read global settings
	if( RegCreateKey( HKEY_LOCAL_MACHINE, m_GlobalRegistryKey, &hKey ) == ERROR_SUCCESS) {
	
		len = sizeof(m_Properties);
		RegQueryValueEx( hKey, _T("PurgeInfo"), NULL, NULL, (unsigned char *) &m_Properties, 
			&len );

		// read extension filters
		numFilters = 0;
		len = sizeof(numFilters);
		RegQueryValueEx( hKey, _T("NumFilters"), NULL, NULL, (unsigned char *) &numFilters, 
			&len );
		if( numFilters ) {

			len = numFilters * sizeof(FILTERENTRY);
			filters = (PFILTERENTRY) malloc( len );
			RegQueryValueEx( hKey, _T("Filter"), NULL, NULL, (unsigned char *) filters, 
				&len );

			// fill in filter array
			for( i = 0; i < numFilters; i++ ) {

				m_ExtensionFilters.AddTail( filters[i].filter );
			}
			free( filters );
		}

		// read directory filters
		numFilters = 0;
		len = sizeof(numFilters);
		RegQueryValueEx( hKey, _T("NumDirFilters"), NULL, NULL, (unsigned char *) &numFilters, 
			&len );
		if( numFilters ) {

			sizeDirFilters = 0;
			len = sizeof(sizeDirFilters);
			RegQueryValueEx( hKey, _T("SizeDirFilters"), NULL, NULL, (unsigned char *) &sizeDirFilters, 
				&len );

			dirFilters = (PCHAR) malloc( sizeDirFilters );
			len = sizeDirFilters;
			RegQueryValueEx( hKey, _T("DirFilter"), NULL, NULL, (unsigned char *) dirFilters, 
				&len );

			// march through extracting strings
			index = 0;
			for( i = 0; i < numFilters; i++ ) {

				filterString = &dirFilters[index];
				m_DirectoryFilters.AddTail( filterString );
				index += filterString.GetLength()+1;
			}
			free( dirFilters );
		}

		RegCloseKey( hKey );
	}

	// tell driver about the mutex (if the driver is there)
	if( (driverHandle = CreateFile( "\\\\.\\UNDELETE",
						  GENERIC_READ | GENERIC_WRITE,
						  0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
						  NULL )) != ((HANDLE)-1) ) {

		// Have driver update the filter
		hMutex = HANDLE(m_Mutex);
		if ( ! DeviceIoControl(	driverHandle, UNDELETE_MUTEX,
								&hMutex, sizeof(HANDLE), 
								NULL, 0, &len, NULL ) ) {

			MessageBox( "Could not update Driver with synchronization information", 
				"Undelete Error", MB_ICONERROR );
		}
		CloseHandle( driverHandle );
	}

	// make default local settings
	m_PromptOnAction = TRUE;

	// read local settings
	if( RegCreateKey( HKEY_CURRENT_USER, m_LocalRegistryKey, &hKey ) == ERROR_SUCCESS) {

		len = sizeof(m_PromptOnAction);
		RegQueryValueEx( hKey, _T("Prompt"), NULL, NULL, (unsigned char *) &m_PromptOnAction, 
					&len );
		RegCloseKey( hKey );
	}
}

/////////////////////////////////////////////////////////////////////////////
// data for the list view control

#define NUM_COLUMNS	5
#define NUM_ITEMS	7

CUndeleteView::~CUndeleteView()
{
	HKEY	hKey;
	DWORD	len;
	DWORD	index;
	DWORD	numFilters;
	PFILTERENTRY filters;
	PCHAR	dirFilters;
	CString	filterString;
	POSITION pos;

	// save settings (this may be redundant, but who cares)
	if( RegOpenKey( HKEY_LOCAL_MACHINE, m_GlobalRegistryKey, &hKey ) == ERROR_SUCCESS) {
	
		RegSetValueEx( hKey, _T("PurgeInfo"), NULL, REG_BINARY, (unsigned char *) &m_Properties, 
			sizeof(m_Properties) );

		// save directory filters
		numFilters = m_DirectoryFilters.GetCount();
		RegSetValueEx( hKey, _T("NumDirFilters"), NULL, REG_DWORD, (unsigned char *) &numFilters, 
			sizeof(numFilters) );

		// calculate size needed
		len = 0;
		for( pos = m_DirectoryFilters.GetHeadPosition(); pos != NULL; m_DirectoryFilters.GetNext(pos)) {

			filterString = m_DirectoryFilters.GetAt(pos);
			len += filterString.GetLength()+1;
		}
		dirFilters = (PCHAR) malloc( len );
		index = 0;
		for( pos = m_DirectoryFilters.GetHeadPosition(); pos != NULL; m_DirectoryFilters.GetNext(pos)) {

			filterString = m_DirectoryFilters.GetAt(pos);
			memcpy( &dirFilters[index], (LPCSTR) filterString, filterString.GetLength()+1);
			index += filterString.GetLength()+1;
		}
		RegSetValueEx( hKey, _T("DirFilter"), NULL, REG_DWORD, (unsigned char *) dirFilters, 
			len );		
		RegSetValueEx( hKey, _T("SizeDirFilters"), NULL, REG_DWORD, (unsigned char *) &len, 
			sizeof(len) );
		free( dirFilters );

		// save exclusion filters
		numFilters = m_ExtensionFilters.GetCount();
		RegSetValueEx( hKey, _T("NumFilters"), NULL, REG_DWORD, (unsigned char *) &numFilters, 
			sizeof(numFilters) );
	
		filters = (PFILTERENTRY) malloc( numFilters * sizeof(FILTERENTRY));
		index = 0;
		for( pos = m_ExtensionFilters.GetHeadPosition(); pos != NULL; m_ExtensionFilters.GetNext(pos)) {

			filterString = m_ExtensionFilters.GetAt(pos);
			strcpy( filters[index++].filter, (LPCSTR) filterString );
		}
		RegSetValueEx( hKey, _T("Filter"), NULL, REG_BINARY, (unsigned char *) filters, 
			numFilters * sizeof(FILTERENTRY) );
		free( filters );
			
		RegCloseKey( hKey );
	}

	// save local settings
	if( RegOpenKey( HKEY_CURRENT_USER, m_LocalRegistryKey, &hKey ) == ERROR_SUCCESS) {

		RegSetValueEx( hKey, _T("Prompt"), NULL, REG_DWORD, (unsigned char *) &m_PromptOnAction, 
					sizeof(m_PromptOnAction) );
		RegCloseKey( hKey );
	}
}

BOOL CUndeleteView::PreCreateWindow(CREATESTRUCT& cs)
{
	cs.style |= LVS_SHOWSELALWAYS | LVS_REPORT;

	return CListViewEx::PreCreateWindow(cs);
}



static _TCHAR *_gszColumnLabel[NUM_COLUMNS] =
{
	_T("Name"), _T("Original Location"), _T("Date Deleted"), _T("Type"),
	_T("Size")
};

static int _gnColumnFmt[NUM_COLUMNS] = 
{
	LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT,
		LVCFMT_RIGHT
};

static int _gnColumnWidth[NUM_COLUMNS] = 
{
	150, 50, 50, 50, 100
};

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView listview fill

void CUndeleteView::ReadBins( BOOLEAN UpdateListView )
{
	int				i;
	DWORD			index, j;
	BOOL			isAnsi = FALSE;
	BOOL			isFat = FALSE;
	_TCHAR			pathString[MAX_PATH];
	_TCHAR			fileString[MAX_PATH];
	DWORD			drives = GetLogicalDrives();
	INFO_HEADER		infoHeader;
	CMenu *hMenu = ((CMainFrame*)AfxGetMainWnd())->GetMenu();
	CToolBar *toolbar = &((CMainFrame*)AfxGetMainWnd())->m_wndToolBar;
	CToolBarCtrl *toolctrl = &toolbar->GetToolBarCtrl();
	CString			menuString;
	UNIINFO_ENTRY	uniEntry;
	ANSIINFO_ENTRY	ansiEntry;
	CFile			infoFile;
	DWORD			fileType;
	ULARGE_INTEGER	freeBytes, totalBytes, totalFreeBytes;
	SHFILEINFO		shellFileInfo;
	CListCtrl&		ListCtrl = GetListCtrl();
	DWORD			readOffset, writeOffset;

	int				totalFiles = 0;
	int				totalSize = 0;

	BeginWaitCursor();
	ListCtrl.SetRedraw( FALSE );

	// clear old contents
	if( UpdateListView ) {

		i = 0;
		while( (index = ListCtrl.GetNextItem( i++, LVNI_ALL )) != -1 ) {

			delete ((CInfoEntry *) ListCtrl.GetItemData( index ));
		}
		ListCtrl.DeleteAllItems();
	}
	memset( m_ValidDriveSize, 0, sizeof(m_ValidDriveSize));

	// grab the big bin mutex
	m_Mutex.Lock();

    SetErrorMode( SEM_FAILCRITICALERRORS ); 

	// get hard drives
	for( i = 2; i < 26; i++ ) {

		_TCHAR	fileSystem[MAX_PATH];
		DWORD	serialNumber, maxComponent, flags;
		_TCHAR	volumeName[MAX_PATH];

		wsprintf( pathString, _T("%C:\\"), i + _T('A'));
		if( drives & (1 << i ) &&
			(GetDriveType( pathString ) == DRIVE_FIXED ) &&
			GetVolumeInformation( pathString, 
							volumeName, sizeof(volumeName), 
							&serialNumber, &maxComponent, &flags, 
							fileSystem, sizeof(fileSystem))) {

			// fill in drive size
			GetDiskFreeSpaceEx( pathString, &freeBytes, &totalBytes,
								&totalFreeBytes );
			m_ValidDriveSize[ i ] = (DWORD) (totalBytes.QuadPart/1024);
			m_ValidDriveLabel[ i ].Format(_T("%s (%C:)"),
							volumeName, i + _T('A' ));

			// open the info file and read the contents
			TRY
			{
				// switch depending on FAT or NTFS
				if( !strnicmp( fileSystem, _T("FAT"), strlen(_T("FAT")))) {

					strcat( pathString, _T("NTRECYCLED\\Info"));
					isFat = TRUE;

				} else {

					wsprintf( pathString, _T("%C:\\NTRecycler\\%s\\Info"),
								i+ _T('A'), m_TextualSid );
					isFat = FALSE;
				}
				if( !infoFile.Open( pathString, 
					CFile::modeReadWrite, NULL )) continue;

				if( infoFile.Read( &infoHeader, sizeof(infoHeader )) !=
					sizeof(infoHeader )) continue;
				isAnsi == (infoHeader.Version == ANSI_RECYCLE_TYPE);
				readOffset = writeOffset = sizeof(infoHeader );
				for( j = 0; j < infoHeader.Entries; j++ ) {
					LV_ITEM lvi;
					CInfoEntry *newEntry;

					// read the entry and add it to the listview
					infoFile.Seek( readOffset, CFile::begin );
					if( isAnsi ) {

						infoFile.Read( &ansiEntry, sizeof(ansiEntry));
						uniEntry = *(PUNIINFO_ENTRY) &ansiEntry;
						readOffset += sizeof(ansiEntry);

					} else {

						infoFile.Read( &uniEntry, sizeof(uniEntry ));
						readOffset += sizeof(uniEntry);
					}

					// Get type information
					if( isFat ) {
						wsprintf( fileString, _T("%C:\\NTRecycled\\D%C%d%s"),
							i+_T('A'), uniEntry.Drive+_T('A'), uniEntry.FileID,
							strrchr( (PCHAR) uniEntry.AsciiName, _T('.')) ? 
							strrchr( (PCHAR) uniEntry.AsciiName, _T('.')) : _T("") );
					} else {
						wsprintf( fileString, _T("%C:\\NTRecycler\\%s\\D%C%d%s"),
							i+_T('A'), m_TextualSid, uniEntry.Drive+_T('A'), uniEntry.FileID,
							strrchr( (PCHAR) uniEntry.AsciiName, _T('.')) ? 
							strrchr( (PCHAR) uniEntry.AsciiName, _T('.')) : _T("") );
					}
					shellFileInfo.szTypeName[0] = 0;
					fileType = SHGetFileInfo( fileString, 0, &shellFileInfo, sizeof(shellFileInfo),
							SHGFI_TYPENAME|SHGFI_EXETYPE );
					if( !shellFileInfo.szTypeName[0] ) {

						if( fileType ) strcpy( shellFileInfo.szTypeName, _T("File"));
						else strcpy( shellFileInfo.szTypeName, _T("Executable Application"));
					}

					if( UpdateListView ) {

						newEntry = new CInfoEntry( uniEntry.FileID, 
													CString( uniEntry.AsciiName ), 
													CString( fileString ),
													CString( shellFileInfo.szTypeName ),
													uniEntry.TimeDeleted, 200 );

						if( newEntry->m_FileId == -1 ) {

							delete newEntry;
							continue;
						}

						// create the entry
						lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_STATE |LVIF_PARAM;
						lvi.iItem = i;
						lvi.iSubItem = 0;
						lvi.pszText = (PCHAR) (LPCTSTR) newEntry->m_Name;
						lvi.iImage = i;
						lvi.lParam = (LONG) newEntry;
						lvi.stateMask = LVIS_STATEIMAGEMASK;
						lvi.state = INDEXTOSTATEIMAGEMASK(1);

						index = ListCtrl.InsertItem(&lvi);

						// set the subitem strings
						ListCtrl.SetItemText(index,1, (LPCSTR) newEntry->m_OriginalName);
						ListCtrl.SetItemText(index,2, (LPCSTR) newEntry->m_TimeDeletedString);
						ListCtrl.SetItemText(index,3, (LPCSTR) newEntry->m_Type);
						ListCtrl.SetItemText(index,4, (LPCSTR) newEntry->m_SizeString);	
						totalFiles++;
						totalSize += newEntry->m_Size;

					} else {

						// does file exist? (was it deleted from the bin)
						CFile deletedFile;

						TRY 
						{
							if( deletedFile.Open( fileString, CFile::modeRead )) {

								// this goes in the bin
								infoFile.Seek( writeOffset, CFile::begin );
								if( isAnsi ) {
									infoFile.Write( &ansiEntry, sizeof(ansiEntry));
									writeOffset += sizeof( ansiEntry );
								} else {
									infoFile.Write( &uniEntry, sizeof(uniEntry));
									writeOffset += sizeof( uniEntry );
								}
								totalFiles++;
								totalSize += deletedFile.GetLength();
							}
						}
						CATCH( CFileException, e )
						{
						}
						END_CATCH
					}
				}

				// update if we're writing a new bin
				if( !UpdateListView ) {
	
					// update the header
					infoHeader.Entries = totalFiles;
					infoFile.Seek( 0, CFile::begin );
					infoFile.Write( &infoHeader, sizeof( infoHeader ));

					// truncate
					infoFile.SetLength( writeOffset );
				}
				infoFile.Close();
			}
			CATCH( CFileException, e )
			{
			
			}
			END_CATCH
		}
	}

    SetErrorMode( 0 ); 

	// unlock the mutex
	m_Mutex.Unlock();

	// update status window
	CString paneText;

	paneText.Format(_T("%d File(s)"), totalFiles );
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 1, paneText, TRUE );

	FormatSizeString( &paneText, totalSize );
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 2, paneText, TRUE );

	// if there are no files, disable the empty menu item and trash can
	if( !totalFiles ) {

		hMenu->GetMenuString( ID_EMPTYBIN, menuString, 0 );
		hMenu->ModifyMenu( ID_EMPTYBIN, 0, -ID_EMPTYBIN, menuString );
		hMenu->EnableMenuItem( -ID_EMPTYBIN, MF_GRAYED );
		toolctrl->SetCmdID( 3, -ID_EMPTYBIN );

	} else {

		hMenu->GetMenuString( -ID_EMPTYBIN, menuString, 0 );
		hMenu->ModifyMenu( -ID_EMPTYBIN, 0, ID_EMPTYBIN, menuString );
		hMenu->EnableMenuItem( ID_EMPTYBIN, MF_ENABLED );
		toolctrl->SetCmdID( 3, ID_EMPTYBIN );
	}

	ListCtrl.SetRedraw( TRUE );
	EndWaitCursor();
}

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView initialization

void CUndeleteView::OnInitialUpdate()
{
	CMenu *hMenu = ((CMainFrame*)AfxGetMainWnd())->GetMenu();
	CString menuString;
	CToolBar *toolbar = &((CMainFrame*)AfxGetMainWnd())->m_wndToolBar;
	CToolBarCtrl *toolctrl = &toolbar->GetToolBarCtrl();

	// read the properties
	CListViewEx::OnInitialUpdate();
	CListCtrl& ListCtrl = GetListCtrl();

	// Gray out menu items that should be disabled when nothing is selected
	// NOTE: This is a hack to get around MFC's helpfulness
	hMenu->GetMenuString( ID_UNDELETE, menuString, 0 );
	hMenu->ModifyMenu( ID_UNDELETE, 0, -ID_UNDELETE, menuString );
	hMenu->EnableMenuItem( -ID_UNDELETE, MF_GRAYED );
	toolctrl->SetCmdID( 0, -ID_UNDELETE );

	hMenu->GetMenuString( ID_DELETE, menuString, 0 );
	hMenu->ModifyMenu( ID_DELETE, 0, -ID_DELETE, menuString );
	hMenu->EnableMenuItem( -ID_DELETE, MF_GRAYED );
	toolctrl->SetCmdID( 1, -ID_DELETE );

	hMenu->GetMenuString( ID_FILEPROPERTIES, menuString, 0 );
	hMenu->ModifyMenu( ID_FILEPROPERTIES, 0, -ID_FILEPROPERTIES, menuString );
	hMenu->EnableMenuItem( -ID_FILEPROPERTIES, MF_GRAYED );

	// set image lists - just to make sizing 
	m_SmallImageList.Create(IDB_SMALLICONS, 16, 1, RGB(255, 255, 255));
	ListCtrl.SetImageList( &m_SmallImageList, LVSIL_SMALL);

	// insert columns
	int i;
	LV_COLUMN lvc;
	CWindowPlacement wp;

	wp.RestoreListView( "Settings", NUM_COLUMNS, this );

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	for(i = 0; i<NUM_COLUMNS; i++)
	{
		lvc.iSubItem = i;
		lvc.pszText = _gszColumnLabel[i];
		lvc.cx = wp.m_ColumnWidth[i];
		lvc.fmt = _gnColumnFmt[i];
		ListCtrl.InsertColumn(i,&lvc);
	}

	// fill the listview
	ReadBins( TRUE );
	SetCompareFunction( CompareFunc );
	Resort(FALSE);
}

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView diagnostics

#ifdef _DEBUG
void CUndeleteView::AssertValid() const
{
	CListViewEx::AssertValid();
}

void CUndeleteView::Dump(CDumpContext& dc) const
{
	CListViewEx::Dump(dc);
}

CUndeleteDoc* CUndeleteView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CUndeleteDoc)));
	return (CUndeleteDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CUndeleteView helpers

BOOL CUndeleteView::SetViewType(DWORD dwViewType)
{
	return(ModifyStyle(LVS_TYPEMASK,dwViewType & LVS_TYPEMASK));
}

DWORD CUndeleteView::GetViewType()
{
	return(GetStyle() & LVS_TYPEMASK);
}



/////////////////////////////////////////////////////////////////////////////
// CUndeleteView message handlers


void CUndeleteView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	UINT uFlags = 0;
	int nHitItem = GetListCtrl().HitTest(point, &uFlags);

	if (uFlags & LVHT_ONITEM)
	{


	}

	CListViewEx::OnLButtonDblClk(nFlags, point);
}

int CALLBACK CUndeleteView::CompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	int	subItem = (int) abs(lParamSort) - 1;
	CInfoEntry *item1, *item2;
	int difference;

	item1 = (CInfoEntry *) lParam1;
	item2 = (CInfoEntry *) lParam2;

	switch( subItem ) {
	case 0:

		difference = item1->m_Name.Compare( item2->m_Name );
		break;

	case 1:

		difference = item1->m_OriginalName.Compare( item2->m_OriginalName );
		break;

	case 2:
		
		difference = (LONG) (item1->m_TimeDeleted.QuadPart - item2->m_TimeDeleted.QuadPart);
		break;

	case 3:

		difference = item1->m_Type.Compare( item2->m_Type );
		break;

	case 4:

		difference = item1->m_Size - item2->m_Size;
		break;
	}

	if( lParamSort > 0 ) return difference;
	else		 return -difference;
}


BOOL CUndeleteView::GetTextualSID( VOID )
{
    PSID_IDENTIFIER_AUTHORITY psia;
    DWORD dwSubAuthorities;
    DWORD dwSidRev=SID_REVISION;
    DWORD dwCounter;
    DWORD dwSidSize;
	DWORD cbBuffer;
	HANDLE hToken;
    PTOKEN_USER ptgUser;  

	// get current token
	if(!OpenProcessToken( GetCurrentProcess(), TOKEN_QUERY, &hToken )) {

		return FALSE;
	}

    // obtain user identified by current process' access token    //
    GetTokenInformation(
					hToken,   
					TokenUser,
					NULL,  
					0, 
					&cbBuffer );
	ptgUser = (PTOKEN_USER) malloc( cbBuffer );

    if( !GetTokenInformation(
					hToken,   
					TokenUser,
					ptgUser,  
					cbBuffer, 
					&cbBuffer ) ) {

		CloseHandle(hToken);
		free( ptgUser );
		return FALSE;
	}

    // close token handle.
	CloseHandle(hToken);

    // test if Sid passed in is valid
    if(!IsValidSid(ptgUser->User.Sid)) return FALSE;

    // obtain SidIdentifierAuthority
    psia=GetSidIdentifierAuthority(ptgUser->User.Sid);

    // obtain sidsubauthority count
    dwSubAuthorities=*GetSidSubAuthorityCount(ptgUser->User.Sid);

    //
    // compute buffer length
    // S-SID_REVISION- + identifierauthority- + subauthorities- + NULL
    //
    dwSidSize=(15 + 12 + (12 * dwSubAuthorities) + 1) * sizeof(TCHAR);

    //
    // check provided buffer length.
    // If not large enough, indicate proper size and setlasterror
    //
    m_TextualSid = (LPTSTR) malloc(dwSidSize);

    //
    // prepare S-SID_REVISION-
    //
    dwSidSize=wsprintf(m_TextualSid, _T("S-%lu-"), dwSidRev );

    //
    // prepare SidIdentifierAuthority
    //
    if ( (psia->Value[0] != 0) || (psia->Value[1] != 0) )
    {
        dwSidSize+=wsprintf(m_TextualSid + lstrlen(m_TextualSid),
                    _T("0x%02hx%02hx%02hx%02hx%02hx%02hx"),
                    (USHORT)psia->Value[0],
                    (USHORT)psia->Value[1],
                    (USHORT)psia->Value[2],
                    (USHORT)psia->Value[3],
                    (USHORT)psia->Value[4],
                    (USHORT)psia->Value[5]);
    }
    else
    {
        dwSidSize+=wsprintf(m_TextualSid + lstrlen(m_TextualSid),
                    _T("%lu"),
                    (ULONG)(psia->Value[5]      )   +
                    (ULONG)(psia->Value[4] <<  8)   +
                    (ULONG)(psia->Value[3] << 16)   +
                    (ULONG)(psia->Value[2] << 24)   );
    }

    //
    // loop through SidSubAuthorities
    //
    for (dwCounter=0 ; dwCounter < dwSubAuthorities ; dwCounter++)
    {
        dwSidSize+=wsprintf(m_TextualSid + dwSidSize, _T("-%lu"),
                    *GetSidSubAuthority(ptgUser->User.Sid, dwCounter) );
    }

	free( ptgUser );
    return TRUE;
}

void CUndeleteView::OnDestroy() 
{
	CListViewEx::OnDestroy();
	
	CWindowPlacement wp;

	wp.SaveListView( "Settings", NUM_COLUMNS, this );
}

VOID CUndeleteView::DeleteSelectedEntries( BOOL Prompt, BOOL DisableRedraw )
{
	int	nItem, prevItem = 0;;
	LV_ITEM listItem;
	CInfoEntry *itemEntry;
	CListCtrl& ListCtrl = GetListCtrl();

	if( Prompt && m_PromptOnAction && MessageBox( _T("Are you sure you want to delete the selected file(s)?"),
		_T("Confirm Delete"), MB_YESNO|MB_ICONQUESTION ) == IDNO ) {

		return;
	}

	BeginWaitCursor();
	if( DisableRedraw ) {
		SetRedraw(FALSE);
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetRedraw(FALSE);
	}

	// update selection 
	nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);

	do {
		itemEntry = (CInfoEntry *) ListCtrl.GetItemData(nItem);
		CFile entryFile;

		// delete the file in the bin - the file may have already been deleted
		// because of space restrictions by the driver, so ignore errors
		TRY 
		{
			entryFile.Remove( itemEntry->m_DeletedName ); 
		} 
		CATCH( CFileException, e ) 
		{
		} 
		END_CATCH;

		ListCtrl.DeleteItem( nItem );
		delete itemEntry;

		prevItem = nItem;
		nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);

	} while	(nItem != -1);

	listItem.mask = LVIF_STATE;
	listItem.iItem = 
		prevItem >= ListCtrl.GetItemCount() ? ListCtrl.GetItemCount()-1 : prevItem;
	listItem.iSubItem = 0;
	listItem.stateMask = listItem.state = LVIS_SELECTED | LVIS_FOCUSED;
	ListCtrl.SetItem( &listItem );
	ListCtrl.EnsureVisible( listItem.iItem, FALSE );
	
	// redraw the listview
	InvalidateRgn( NULL );

	EndWaitCursor();

	// if not prompting, we have to update the bin
	if( Prompt ) ReadBins( FALSE );

	if( DisableRedraw) {
		SetRedraw(TRUE);
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetRedraw(TRUE);
	}
}


VOID CUndeleteView::ActionSelectedEntries()
{
	static BOOL FirstTime = TRUE;
	int	nItem;
	int index, length;
	int numDeleted = 0;
	CInfoEntry *itemEntry;
	CListCtrl& ListCtrl = GetListCtrl();
	CString dirName, remainDir;
	CFileException eFile;
	BOOL notDeleted = FALSE, notFound = FALSE;

	if( m_TotalSelected == 0 ) {

		return;
	}

	if( m_PromptOnAction &&
		MessageBox( _T("Restore selected file(s)?"),
			_T("Confirm Restore"), MB_YESNO|MB_ICONQUESTION ) == IDNO ) {
	
		return;
	}

	BeginWaitCursor();
	SetRedraw( FALSE );

	// update selection 
	nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);

	do {
		itemEntry = (CInfoEntry *) ListCtrl.GetItemData(nItem);
		CFile entryFile;
		CFile checkFile;

		// make sure original doesn't exist
		if( checkFile.Open( itemEntry->m_OriginalName, CFile::modeRead, &eFile )) {

			notDeleted = TRUE;
			nItem = ListCtrl.GetNextItem(nItem, LVNI_SELECTED);
			checkFile.Close();
			continue;
		}		

		// create the target directory
		index = 4; // skip 'X:\'
		length = itemEntry->m_OriginalName.GetLength();
		while( index < length) {

			if( itemEntry->m_OriginalName[index] == '\\' ) {

				dirName = itemEntry->m_OriginalName.Left( index );
				_mkdir( dirName );
			}
			index++;
		}	
		
		// take the file out of the bin - note that the file may not be
		// in the bin anymore if the driver was making space and deleted it
		TRY
		{
			entryFile.Rename( itemEntry->m_DeletedName, itemEntry->m_OriginalName );
		}
		CATCH( CFileException, e )
		{
			notFound = TRUE;
		}
		END_CATCH

		ListCtrl.DeleteItem( nItem );
		delete itemEntry;
		numDeleted++;
		nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);

	} while	(nItem != -1);

	if( numDeleted ) {

		// redraw the listview
		InvalidateRgn( NULL );
	}
	EndWaitCursor();
	SetRedraw(TRUE);

	if( notDeleted ) {

		MessageBox( _T("Existing versions of one or more files were found ")
					_T("and will not be overwritten.\nUndelete files that ")
					_T("match existing file names will remain in the Bin."),
				_T("Undelete"), MB_OK|MB_ICONINFORMATION );
	}


	if( notFound ) {

		MessageBox( _T("One or more selected files were not restored because ")
					_T("they had already been removed from the bin."),
				_T("Undelete"), MB_OK|MB_ICONWARNING );
	}

	ReadBins( FALSE );
	UpdateStatusBar();
	
	if( numDeleted && FirstTime ) {

		CWinternals Winternals;
		Winternals.DoModal();
		FirstTime = FALSE;
	}
}

void CUndeleteView::OnDelete() 
{

	if( m_TotalSelected == 0 ) {

		MessageBox( _T("No files selected to delete."),
			_T("Nothing Selected"), MB_OK|MB_ICONINFORMATION );
		return;
	}

	if( !m_PromptOnAction || MessageBox( _T("Are you sure you want to delete the selected file(s)?"),
		_T("Confirm Delete"), MB_YESNO|MB_ICONQUESTION ) == IDYES ) {
	
		DeleteSelectedEntries( FALSE, TRUE );

		// now rebuild the bin
		ReadBins( FALSE );
		UpdateStatusBar();
	}	
}

void CUndeleteView::OnEmptybin() 
{
	int	nItem;
	LV_ITEM listItem;
	CInfoEntry *itemEntry;
	CListCtrl& ListCtrl = GetListCtrl();
	CString paneText;

	if( m_PromptOnAction && 
		MessageBox( _T("Are you sure you want to delete ALL existing files from the Undelete Bin?"),
			_T("Confirm Delete"), MB_YESNO|MB_ICONQUESTION ) == IDNO ) {
	
		return;
	}	

	SetRedraw( FALSE);
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetRedraw(FALSE);

	// update selection 
	m_TotalSelected = 0;
	nItem = ListCtrl.GetNextItem(-1, LVNI_ALL);

	while( nItem != -1 ) {

		itemEntry = (CInfoEntry *) ListCtrl.GetItemData(nItem);

		listItem.mask = LVIF_STATE;
		listItem.iItem = nItem;
		listItem.iSubItem = 0;
		listItem.stateMask = listItem.state = LVIS_SELECTED;
		ListCtrl.SetItem( &listItem );

		nItem = ListCtrl.GetNextItem( nItem, LVNI_ALL );
		m_TotalSelected++;
	} 

	if( m_TotalSelected ) {
		
		DeleteSelectedEntries( FALSE, FALSE );
		ReadBins( FALSE );
		m_TotalSelected = 0;
		SelectionChanged( 0 );
	}

	SetRedraw(TRUE);
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetRedraw(TRUE);

	paneText =  _T("0 File(s)");
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 1, paneText, TRUE );

	paneText = _T("0KB");
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 2, paneText, TRUE );

	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.InvalidateRgn( NULL );
}

void CUndeleteView::OnUndelete() 
{
	if( m_TotalSelected == 0 ) {

		MessageBox( _T("No files selected to restore."),
			_T("Nothing Selected"), MB_OK|MB_ICONINFORMATION );
		return;
	}
	ActionSelectedEntries();	
}

VOID CUndeleteView::SelectionChanged(DWORD TotalSelected)
{
	CMenu *hMenu = ((CMainFrame*)AfxGetMainWnd())->GetMenu();
	CString menuString;
	CToolBar *toolbar = &((CMainFrame*)AfxGetMainWnd())->m_wndToolBar;
	CToolBarCtrl *toolctrl = &toolbar->GetToolBarCtrl();

	// need to update the menu selections - they should only be 
	// enabled when something is selected
	if( !TotalSelected ) {

		hMenu->GetMenuString( ID_UNDELETE, menuString, 0 );
		hMenu->ModifyMenu( ID_UNDELETE, 0, -ID_UNDELETE, menuString );
		hMenu->EnableMenuItem( -ID_UNDELETE, MF_GRAYED );
		toolctrl->SetCmdID( 0, -ID_UNDELETE );

		hMenu->GetMenuString( ID_DELETE, menuString, 0 );
		hMenu->ModifyMenu( ID_DELETE, 0, -ID_DELETE, menuString );
		hMenu->EnableMenuItem( -ID_DELETE, MF_GRAYED );
		toolctrl->SetCmdID( 1, -ID_DELETE );

		hMenu->GetMenuString( ID_FILEPROPERTIES, menuString, 0 );
		hMenu->ModifyMenu( ID_FILEPROPERTIES, 0, -ID_FILEPROPERTIES, menuString );
		hMenu->EnableMenuItem( -ID_FILEPROPERTIES, MF_GRAYED );

	} else {

		hMenu->GetMenuString( -ID_UNDELETE, menuString, 0 );
		hMenu->ModifyMenu( -ID_UNDELETE, 0, ID_UNDELETE, menuString );
		hMenu->EnableMenuItem( ID_UNDELETE, MF_ENABLED );
		toolctrl->SetCmdID( 0, ID_UNDELETE );

		hMenu->GetMenuString( -ID_DELETE, menuString, 0 );
		hMenu->ModifyMenu( -ID_DELETE, 0, ID_DELETE, menuString );
		hMenu->EnableMenuItem( ID_DELETE, MF_ENABLED );
		toolctrl->SetCmdID( 1, ID_DELETE );

		hMenu->GetMenuString( -ID_FILEPROPERTIES, menuString, 0 );
		hMenu->ModifyMenu( -ID_FILEPROPERTIES, 0, ID_FILEPROPERTIES, menuString );
		hMenu->EnableMenuItem( ID_FILEPROPERTIES, MF_ENABLED );
	}
}

void CUndeleteView::OnRefresh() 
{
	// Redo the list view
	ReadBins( TRUE );	
	Resort(FALSE);
}

void CUndeleteView::OnProperties() 
{
	int		i;
	HKEY hKey;
	DWORD len, type;
	DWORD shellState[256];

	CPropSh propSheet(_T("Undelete Properties"));
	CProp  mainPage;
	CDriveProp *newPage;

	// fill in main page
	propSheet.SetCallbackWindow( this );

	mainPage.m_ConfigureGlobal = m_Properties.EnforceGlobalSetting;
	mainPage.m_PromptOnAction = m_PromptOnAction;
	mainPage.m_immediatedelete = 
		m_Properties.RecycleDisableMask & (1 << GLOBAL_SETTING) ? 1 : 0;
	mainPage.m_GlobalSize = m_Properties.ReservedForRecycle[GLOBAL_SETTING];

	// get explorer delete flag
	if( RegOpenKey( HKEY_CURRENT_USER, _T("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer"), &hKey ) == ERROR_SUCCESS) {
	
		len = sizeof(shellState);
		RegQueryValueEx( hKey, _T("ShellState"), NULL, &type, (unsigned char *) shellState, 
			&len );
		RegCloseKey( hKey );
		mainPage.m_ExpPrompt = (shellState[1] & SHELLPROMPTFLAG ? 0: 1);

		// changing the properties doesn't work on older versions of the shell
		mainPage.m_DisableExpPrompt = FALSE;
		if( len <= 16 ) {

			mainPage.m_DisableExpPrompt = TRUE;
		} 
	}	

	propSheet.m_GlobalPropertyPage = &mainPage;
	propSheet.AddPage( &mainPage );

	// create a page for each drive
	for( i = 0; i < 26; i++ ) {

		propSheet.m_DrivePropertyPages[i] = NULL;
		if( m_ValidDriveSize[i] ) {

			newPage = new CDriveProp;

			newPage->m_psp.dwFlags |= PSP_USETITLE;
			newPage->m_psp.pszTitle = m_ValidDriveLabel[i];
			newPage->m_SizeOfDrive = m_ValidDriveSize[i];
			newPage->m_GlobalProperties = &mainPage;
			newPage->m_RecycleSize = m_Properties.ReservedForRecycle[i];
			newPage->m_ImmediateDelete = 
				(m_Properties.RecycleDisableMask & (1 << i)) ? 1 : 0;
			propSheet.m_DrivePropertyPages[i] = newPage;
			propSheet.AddPage( newPage );

		}
	}
	propSheet.DoModal();
}

void CUndeleteView::OnExclude() 
{
	CFilter filterDialog;

	filterDialog.m_ValidDriveSize = m_ValidDriveSize;
	filterDialog.m_DirectoryFilters = &m_DirectoryFilters;
	filterDialog.m_ExtensionFilters = &m_ExtensionFilters;
	filterDialog.DoModal();
}

VOID CUndeleteView::ApplyGlobal(CPropSh * pPropSheet)
{
	HKEY hKey;
	DWORD shellState[256];
	DWORD len, type;
	int	i;
	CProp * pGlobalPage = pPropSheet->m_GlobalPropertyPage;

	// get global settings
	m_Properties.EnforceGlobalSetting = pGlobalPage->m_ConfigureGlobal;
	m_Properties.ReservedForRecycle[GLOBAL_SETTING] = pGlobalPage->m_percentrecycle.GetPos();
	m_Properties.RecycleDisableMask = 0;
	m_Properties.RecycleDisableMask |= 
			(pGlobalPage->m_immediatedelete ? 1 : 0) << GLOBAL_SETTING;
	m_PromptOnAction = pGlobalPage->m_PromptOnAction;

	// get explorer delete flag
	if( RegOpenKey( HKEY_CURRENT_USER, _T("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer"), &hKey ) == ERROR_SUCCESS) {
	
		len = sizeof(shellState);
		RegQueryValueEx( hKey, _T("ShellState"), NULL, &type, (unsigned char *) shellState, 
			&len );
		shellState[1] &= ~SHELLPROMPTFLAG;
		shellState[1] |= (pGlobalPage->m_ExpPrompt ?  0 : SHELLPROMPTFLAG);
		RegSetValueEx( hKey, _T("ShellState"), NULL, type, (unsigned char *) shellState, 
			len );
		RegCloseKey( hKey );
	}
	pGlobalPage->SetModified( FALSE );

	// now save local settings
	for( i = 0; i < 26; i++ ) {

		if( m_ValidDriveSize[i] ) {

			m_Properties.ReservedForRecycle[i] = 
				pPropSheet->m_DrivePropertyPages[i]->m_RecycleSize;
			m_Properties.RecycleDisableMask |= 
				(pPropSheet->m_DrivePropertyPages[i]->m_ImmediateDelete ? 1 : 0) << i;
			pPropSheet->m_DrivePropertyPages[i]->SetModified( FALSE );
		}
	}

	// save settings 
	if( RegOpenKey( HKEY_LOCAL_MACHINE, m_GlobalRegistryKey, &hKey ) == ERROR_SUCCESS) {
	
		RegSetValueEx( hKey, _T("PurgeInfo"), NULL, REG_BINARY, (unsigned char *) &m_Properties, 
			sizeof(m_Properties) );
		RegCloseKey( hKey );
	}	
}

void CUndeleteView::OnContextMenu(CWnd* pWnd, CPoint point) 
{
	CMenu	menu;
	BOOL	selected = FALSE;
	int	nItem;
	CListCtrl& ListCtrl = GetListCtrl();
	
	menu.LoadMenu(IDR_CONTEXT);
	CMenu *pContextMenu = menu.GetSubMenu(0);

	// see if any items are selected
	for(nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);
		nItem != -1; nItem = ListCtrl.GetNextItem(nItem, LVNI_SELECTED))
	{
		selected = TRUE;
		break;
	}

	if( !selected ) {

		pContextMenu->EnableMenuItem( ID_DELETE, MF_GRAYED );
		pContextMenu->EnableMenuItem( ID_UNDELETE, MF_GRAYED );
		pContextMenu->EnableMenuItem( ID_FILEPROPERTIES, MF_GRAYED );
	}

	pContextMenu->TrackPopupMenu( TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_RIGHTBUTTON,
		point.x, point.y, this );
}

void CUndeleteView::OnActivate(UINT nState, CWnd* pWndOther, BOOL bMinimized) 
{
	CListViewEx::OnActivate(nState, pWndOther, bMinimized);
	
	// TODO: Add your message handler code here
	
}


VOID CUndeleteView::UpdateStatusBar()
{
	CListCtrl&		ListCtrl = GetListCtrl();
	DWORD totalSelected, totalSelectedSize;
	DWORD totalFiles, totalSize;
	int nItem;
	CInfoEntry *itemEntry;
	CString paneText;

	// reset the status bar
	totalSelected = 0;
	totalSelectedSize = 0;
	totalFiles = 0;
	totalSize = 0;

	for(nItem = ListCtrl.GetNextItem(-1, LVNI_ALL);
		nItem != -1; nItem = ListCtrl.GetNextItem(nItem, LVNI_ALL))
	{
		itemEntry = (CInfoEntry *) ListCtrl.GetItemData(nItem);

		if( ListCtrl.GetItemState(nItem, LVIS_SELECTED )) {

			totalSelected++;
			totalSelectedSize += itemEntry->m_Size;			
		} 
		totalSize += itemEntry->m_Size;	
		totalFiles++;		
	}

	if( totalSelected ) {

		paneText.Format(_T("%d File(s) Selected"), totalSelected );
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 1, paneText, TRUE );

		FormatSizeString( &paneText, totalSelectedSize );
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 2, paneText, TRUE );

	} else {

		paneText.Format(_T("%d File(s)"), totalFiles );
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 1, paneText, TRUE );

		FormatSizeString( &paneText, totalSize );
		((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.SetPaneText( 2, paneText, TRUE );
	}
	((CMainFrame*)AfxGetMainWnd())->m_wndStatusBar.InvalidateRgn( NULL );
}

VOID CUndeleteView::OnFileProperties()
{
	CListCtrl&		ListCtrl = GetListCtrl();
	CInfoEntry	*itemEntry;
	CString		title;
	CFileProp	fileProp;
	CPropertySheet propSheet;
	int nItem;

	// get first item selected
	nItem = ListCtrl.GetNextItem(-1, LVNI_SELECTED);
	if( nItem == -1 ) return;

	itemEntry = (CInfoEntry *) ListCtrl.GetItemData(nItem);
	fileProp.m_InfoEntry = itemEntry;

	title.Format(_T("%s Properties"), itemEntry->m_Name );
	propSheet.SetTitle( title );
	propSheet.AddPage( &fileProp );
	propSheet.DoModal();
}

void CUndeleteView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if( nChar == VK_RETURN && m_TotalSelected ) {

		OnFileProperties();
	}
	CListViewEx::OnKeyDown(nChar, nRepCnt, nFlags);
}

